from typing import Union
import torch


def apply_mask(
    values: torch.FloatTensor, mask: Union[torch.BoolTensor, torch.IntTensor, torch.FloatTensor]
) -> torch.FloatTensor:
    """
    # Parameters

    values : `torch.FloatTensor`
        Shape `batch_size x num_attention_heads x source_seq_len x target_seq_len`
    mask : `torch.BoolTensor`
        Shape `batch_size x target_seq_len` OR `batch_size x 1 x 1 x target_seq_len`
    """
    if len(mask.shape) == 2:
        # We create a 4D attention mask from a 2D tensor mask.
        # The shape is `batch_size x 1 x 1 x target_seq_len` which is broadcast
        # to `batch_size x num_attention_heads x source_seq_len x target_seq_len`
        mask = mask.unsqueeze(1).unsqueeze(2)
    # `mask==1` to convert float tensors.
    mask = (
        ~(mask == 1)
    ) * -10e5  # -10e5 to ensure that the model also works in half-precision mode.
    return values + mask



from typing import Union, Optional, Dict
import logging

import torch

from allennlp.common import FromParams

from allennlp.modules.util import replicate_layers
from allennlp.modules.transformer.transformer_layer import TransformerLayer
from allennlp.modules.transformer.transformer_module import TransformerModule

logger = logging.getLogger(__name__)


class TransformerStack(TransformerModule, FromParams):
    """
    This module is the basic transformer stack.
    Details in the paper:
    [BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding, Devlin et al, 2019]
    (https://api.semanticscholar.org/CorpusID:52967399)

    # Parameters

    num_hidden_layers : `int`
    layer : `TransformerLayer`, optional
    hidden_size : `int`, optional
        This needs to be provided if no `layer` argument is passed.
    intermediate_size : `int`, optional
        This needs to be provided if no `layer` argument is passed.
    num_attention_heads : `int`
    attention_dropout : `float` (default = `0.0`)
        Dropout probability for the `SelfAttention` layer.
    hidden_dropout : `float` (default = `0.0`)
        Dropout probability for the `OutputLayer`.
    activation : `Union[str, torch.nn.Module]` (default = `"relu"`)
    add_cross_attention: `bool` (default = `False`)
        If True, the `TransformerLayer` modules will have cross attention modules as well.
        This is helpful when using the `TransformerStack` as a decoder.
    """

    _huggingface_mapping = {"layer": "layers"}
    _relevant_module = "encoder"

    def __init__(
        self,
        num_hidden_layers: int,
        layer: Optional[TransformerLayer] = None,
        hidden_size: Optional[int] = None,
        intermediate_size: Optional[int] = None,
        num_attention_heads: int = 8,
        attention_dropout: float = 0.1,
        hidden_dropout: float = 0.1,
        activation: Union[str, torch.nn.Module] = "relu",
        add_cross_attention: bool = False,
    ):
        super().__init__()

        if layer is not None:
            logger.warning(
                "The `layer` argument has been specified. Any other arguments will be ignored."
            )
        else:
            assert (hidden_size is not None) and (intermediate_size is not None), "As the `layer`"
            "has not been provided, `hidden_size` and `intermediate_size` are"
            "required to create `TransformerLayer`s."

        layer = layer or TransformerLayer(
            hidden_size,  # type: ignore
            intermediate_size,  # type: ignore
            num_attention_heads,
            attention_dropout,
            hidden_dropout,
            activation,
            add_cross_attention,
        )
        self.layers = replicate_layers(layer, num_hidden_layers)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        output_attentions: bool = False,
        output_hidden_states: bool = False,
    ):
        """
        hidden_states : `torch.Tensor`
            Shape `batch_size x seq_len x hidden_dim`
        attention_mask : `torch.BoolTensor`, optional
            Shape `batch_size x seq_len`
        head_mask : `torch.BoolTensor`, optional
        output_attentions : `bool`
            Whether to also return the attention probabilities, default = `False`
        output_hidden_states : `bool`
            Whether to return the hidden_states for all layers, default = `False`
        """
        all_hidden_states = () if output_hidden_states else None
        all_attentions = () if output_attentions else None
        all_cross_attentions = () if output_attentions and self._add_cross_attention else None
        for i, layer_module in enumerate(self.layers):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)  # type: ignore

            layer_head_mask = head_mask[i] if head_mask is not None else None

            layer_outputs = layer_module(
                hidden_states,
                attention_mask,
                layer_head_mask,
                encoder_hidden_states,
                encoder_attention_mask,
                output_attentions,
            )
            hidden_states = layer_outputs[0]
            if output_attentions:
                all_attentions = all_attentions + (layer_outputs[1],)  # type: ignore
                if self._add_cross_attention:
                    all_cross_attentions = all_cross_attentions + (layer_outputs[2],)  # type: ignore

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)  # type: ignore

        return tuple(
            v
            for v in [hidden_states, all_hidden_states, all_attentions, all_cross_attentions]
            if v is not None
        )

    @classmethod
    def _get_input_arguments(
        cls,
        pretrained_module: torch.nn.Module,
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        submodules = cls._get_mapped_submodules(pretrained_module, source, mapping)

        final_kwargs = {}

        final_kwargs["num_hidden_layers"] = len(submodules["layers"])

        final_kwargs["hidden_size"] = submodules["layers.0.attention.self.query"].in_features
        final_kwargs["num_attention_heads"] = submodules[
            "layers.0.attention.self"
        ].num_attention_heads
        final_kwargs["attention_dropout"] = submodules["layers.0.attention.self.dropout"].p
        final_kwargs["hidden_dropout"] = submodules["layers.0.attention.output.dropout"].p
        final_kwargs["intermediate_size"] = submodules["layers.0.intermediate.dense"].out_features

        # We require the if block as `act_fn` is a function rather than a module,
        # so `_get_mapped_submodules` does not automatically fix this.
        if source == "huggingface":
            final_kwargs["activation"] = getattr(
                submodules["layers.0.intermediate"], "intermediate_act_fn"
            )
        else:
            final_kwargs["activation"] = getattr(submodules["layers.0.intermediate"], "act_fn")

        final_kwargs["add_cross_attention"] = "layers.0.cross_attention" in submodules

        final_kwargs.update(**kwargs)

        return final_kwargs

    @classmethod
    def from_pretrained_module(  # type: ignore
        cls,
        pretrained_module: Union[str, torch.nn.Module],
        num_hidden_layers: Optional[Union[int, range]] = None,
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        final_kwargs = {}
        if num_hidden_layers is not None:
            if isinstance(num_hidden_layers, range):
                if mapping is None:
                    mapping = {}
                    for num_layer, mapped in enumerate(num_hidden_layers):
                        mapping[str(mapped)] = str(num_layer)
                final_kwargs["num_hidden_layers"] = len(num_hidden_layers)
            else:
                final_kwargs["num_hidden_layers"] = num_hidden_layers

        return super().from_pretrained_module(pretrained_module, source, mapping, **final_kwargs)



from typing import Optional, Dict
import torch

from allennlp.common import FromParams
from allennlp.modules.attention import Attention
from allennlp.modules.transformer.transformer_module import TransformerModule
from allennlp.modules.transformer.util import apply_mask


class SelfAttention(TransformerModule, FromParams):
    """
    This module computes the self-attention, similar to the architecture in BERT. Additionally, the attention
    scoring function can be specified.
    Details in the paper:
    [BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding, Devlin et al, 2019]
    (https://api.semanticscholar.org/CorpusID:52967399)

    # Parameters

    hidden_size: `int`
    num_attention_heads: `int`
    dropout: `float` (default = `0.0`)
    scoring_func: `str` (default = `scaled_dot_product`)
        The name of the attention-calculating function to be used.
        Eg. `additive`, `linear`, etc. For a complete list, please check :mod:`allennlp.modules.attention`.
    """

    _relevant_module = ["encoder.layers.0.attention.self", "encoder.layers.0.attention"]
    _huggingface_mapping = {"layer": "layers"}

    def __init__(
        self,
        hidden_size: int,
        num_attention_heads: int,
        dropout: float = 0.0,
        scoring_func: str = "scaled_dot_product",
        output_linear: bool = False,
    ):
        super().__init__()
        if hidden_size % num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (hidden_size, num_attention_heads)
            )
        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads
        self.attention_head_size = int(hidden_size / num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = torch.nn.Linear(hidden_size, self.all_head_size)
        self.key = torch.nn.Linear(hidden_size, self.all_head_size)
        self.value = torch.nn.Linear(hidden_size, self.all_head_size)

        self.scoring_func = scoring_func
        if self.scoring_func in ["additive", "linear", "bilinear"]:
            self.attn = Attention.by_name(self.scoring_func)(hidden_size, hidden_size)
        elif self.scoring_func == "scaled_dot_product":
            self.attn = Attention.by_name(self.scoring_func)(self.attention_head_size, False)
        else:
            self.attn = Attention.by_name(self.scoring_func)()

        # out linear layer for distilbert.
        if output_linear:
            self.output = torch.nn.Linear(hidden_size, self.all_head_size)

        self.dropout = torch.nn.Dropout(dropout)

    def _transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (
            self.num_attention_heads,
            self.attention_head_size,
        )
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(
        self,
        query_states: torch.Tensor,
        key_states: Optional[torch.Tensor] = None,
        value_states: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.BoolTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        output_attentions: bool = False,
    ):
        """
        query_states : `torch.Tensor`
            Shape `batch_size x seq_len x hidden_dim`
        key_states : `torch.Tensor`, optional
            Shape `batch_size x seq_len x hidden_dim`
        value_states : `torch.Tensor`, optional
            Shape `batch_size x seq_len x hidden_dim`
        attention_mask : `torch.BoolTensor`, optional
            Shape `batch_size x seq_len`
        head_mask : `torch.BoolTensor`, optional
        output_attentions : `bool`
            Whether to also return the attention probabilities, default = `False`
        """
        if key_states is None:
            key_states = query_states
        if value_states is None:
            value_states = query_states

        mixed_query_layer = self.query(query_states)
        mixed_key_layer = self.key(key_states)
        mixed_value_layer = self.value(value_states)

        query_layer = self._transpose_for_scores(mixed_query_layer)
        key_layer = self._transpose_for_scores(mixed_key_layer)
        value_layer = self._transpose_for_scores(mixed_value_layer)

        attention_scores = self.attn(query_layer, key_layer.transpose(-1, -2))

        if attention_mask is not None:
            attention_scores = apply_mask(attention_scores, attention_mask)

        attention_probs = torch.nn.Softmax(dim=-1)(attention_scores)
        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)

        if hasattr(self, "output"):
            context_layer = self.output(context_layer)

        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)
        return outputs

    @classmethod
    def _get_mapping(
        cls, pretrained_module=None, source="huggingface", mapping: Optional[Dict[str, str]] = None
    ):
        combined_mapping = {}
        if "huggingface" in source:
            combined_mapping.update(cls._huggingface_mapping)
        if mapping is not None:
            combined_mapping.update(mapping)
        if pretrained_module is not None:
            for name, _ in pretrained_module.named_modules():
                if "q_lin" in name:
                    combined_mapping["q_lin"] = "query"
                    combined_mapping["k_lin"] = "key"
                    combined_mapping["v_lin"] = "value"
                    combined_mapping["out_lin"] = "output"
                    combined_mapping["transformer"] = "encoder"
                    break
        return combined_mapping

    @classmethod
    def _get_input_arguments(
        cls,
        pretrained_module: torch.nn.Module,
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        submodules = cls._get_mapped_submodules(pretrained_module, source, mapping)
        final_kwargs = {}

        final_kwargs["hidden_size"] = submodules["query"].in_features
        if hasattr(submodules[""], "num_attention_heads"):
            final_kwargs["num_attention_heads"] = submodules[""].num_attention_heads
        elif hasattr(submodules[""], "n_heads"):
            final_kwargs["num_attention_heads"] = submodules[""].n_heads
            final_kwargs["output_linear"] = True  # Since this is the distilbert case.
        else:
            raise AttributeError("Cannot find a relevant attribute for number of heads.")

        final_kwargs["dropout"] = submodules["dropout"].p

        final_kwargs.update(**kwargs)

        return final_kwargs



from allennlp.training.callbacks.callback import TrainerCallback
from allennlp.training.callbacks.console_logger import ConsoleLoggerCallback
from allennlp.training.callbacks.sanity_checks import SanityChecksCallback
from allennlp.training.callbacks.tensorboard import TensorBoardCallback
from allennlp.training.callbacks.track_epoch import TrackEpochCallback
from allennlp.training.callbacks.wandb import WandBCallback



from typing import Optional, Dict, Union, List
import logging
import inspect

import torch

from allennlp.common import cached_transformers

logger = logging.getLogger(__name__)


class TransformerModule(torch.nn.Module):
    """
    Base class to help with generalized loading of pretrained weights.

    `_huggingface_mapping` is an optional mapping for each class, that determines
    any differences in the module names between the class modules and the huggingface model's
    modules.

    `_relevant_module` is an optional str or list of str which contains the expected name of the module
    in the huggingface pretrained model. It can be a list to account for different names in different
    models. The search is carried out in the order of the list.
    """

    _huggingface_mapping: Dict[str, str] = {}
    _relevant_module: Optional[Union[str, List[str]]] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _get_mapping(
        cls,
        pretrained_module: Optional[torch.nn.Module] = None,
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
    ):
        """
        Returns the mapping to be used, based on the optional `pretrained_module`.
        If `pretrained_module` is not given, the default module-level mapping is returned.
        """
        combined_mapping = {}
        if "huggingface" in source:
            combined_mapping.update(cls._huggingface_mapping)
        if mapping is not None:
            combined_mapping.update(mapping)
        return combined_mapping

    @classmethod
    def _get_mapped_submodules(
        cls, pretrained_module, source="huggingface", mapping: Optional[Dict[str, str]] = None
    ):
        """
        Subclasses overload this method, and provide appropriate name mapping based on the source.
        """
        submodules = dict(pretrained_module.named_modules())
        combined_mapping = cls._get_mapping(pretrained_module, source, mapping)
        for name, module in pretrained_module.named_modules():
            newname = name
            for key, val in combined_mapping.items():
                newname = newname.replace(key, val)
            submodules[newname] = submodules.pop(name)
        return submodules

    def _construct_default_mapping(
        self,
        pretrained_module,
        source: str = "huggingface",
        mapping: Optional[Dict[str, str]] = None,
    ):
        """
        Recursively constructs the default mapping of parameter names for loading pretrained module weights.
        Keys are parameter names from this module, and values are corresponding parameter names in the
        expected pretrained module, as per `source`.
        """
        combined_mapping = self._get_mapping(pretrained_module, source, mapping)
        for name, module in self.named_modules():
            if name != "":
                if hasattr(module, "_construct_default_mapping"):
                    # We handle collisions by giving priority to the outer module's mapping.
                    combined_mapping = dict(
                        list(
                            module._construct_default_mapping(
                                pretrained_module, source, combined_mapping
                            ).items()
                        )
                        + list(combined_mapping.items())
                    )
        return combined_mapping

    def _load_from_pretrained_module(
        self,
        pretrained_module: torch.nn.Module,
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
        ignore_absent_parameters: Optional[List] = None,
    ):
        """
        Loads the weights of the `pretrained_module` into the instance.
        Optionally, a `mapping` is specified for any differences in parameter names
        between `pretrained_module` and the instance.
        """
        ignore_absent_parameters = ignore_absent_parameters or []
        combined_mapping = self._construct_default_mapping(pretrained_module, source, mapping)
        if mapping is not None:
            combined_mapping.update(mapping)

        inverse_mapping = {val: key for key, val in combined_mapping.items()}
        pretrained_parameters = dict(pretrained_module.named_parameters())
        for name, parameter in self.named_parameters():
            pretrained_name = name
            for key, val in inverse_mapping.items():
                # so that we replace the names of submodules too.
                # eg. module.key.anothermodule --> module.val.anothermodule
                pretrained_name = pretrained_name.replace(key, val)
            if not any(
                [pretrained_name.startswith(paraname) for paraname in ignore_absent_parameters]
            ):
                if pretrained_name not in pretrained_parameters:
                    raise ValueError(
                        f"Couldn't find a matching parameter for {name}. Is this module "
                        "compatible with the pretrained module you're using?"
                    )
                parameter.data.copy_(pretrained_parameters[pretrained_name].data)

    @classmethod
    def _get_input_arguments(
        cls,
        pretrained_module: torch.nn.Module,
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        """
        Constructs the arguments required for instantiating an object of this class, using
        the values from `pretrained_module`.
        """
        return kwargs

    @classmethod
    def get_relevant_module(
        cls,
        pretrained_module: Union[str, torch.nn.Module],
        relevant_module: Optional[Union[str, List[str]]] = None,
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
    ):
        """
        Returns the relevant underlying module given a model name/object.

        # Parameters:

        pretrained_module: Name of the transformer model containing the layer,
                           or the actual layer (not the model object).
        relevant_module: Name of the desired module. Defaults to cls._relevant_module.
        source: Where the model came from. Default - huggingface.
        mapping: Optional mapping that determines any differences in the module names
        between the class modules and the input model's modules. Default - cls._huggingface_mapping
        """
        if isinstance(pretrained_module, str):
            pretrained_module = cached_transformers.get(pretrained_module, False)

        relevant_module = relevant_module or cls._relevant_module

        if relevant_module is not None:
            submodules = cls._get_mapped_submodules(pretrained_module, source, mapping)
            # If the relevant_module is not found, we assume that the pretrained_module
            # is already the relevant module.
            if isinstance(relevant_module, str):
                relevant_module = [relevant_module]
            found = False
            for module in relevant_module:
                if module in submodules:
                    pretrained_module = submodules[module]
                    found = True
                    break

            if not found:
                logger.warning(
                    "{} was not found! The submodules are: {}".format(
                        relevant_module, submodules.keys()
                    )
                )
        return pretrained_module

    @classmethod
    def from_pretrained_module(
        cls,
        pretrained_module: Union[str, torch.nn.Module],
        source="huggingface",
        mapping: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        """
        Creates and returns an instance of the class, by using the weights
        (and the architecture, by default) of the `pretrained_module`.
        Optionally, the architecture can be changed by providing arguments.
        """
        accepted_args = inspect.getfullargspec(cls).args
        accepted_args.remove("self")
        for key in kwargs:
            assert key in accepted_args, (
                "{} is not a valid argument for creating an instance of `{}`. "
                "Accepted arguments are {}.".format(key, cls.__name__, accepted_args)
            )

        pretrained_module = cls.get_relevant_module(
            pretrained_module, source=source, mapping=mapping
        )
        final_kwargs = cls._get_input_arguments(pretrained_module, source, mapping)
        final_kwargs.update(kwargs)
        module = cls(**final_kwargs)
        module._load_from_pretrained_module(pretrained_module, source, mapping)
        return module



import os
from typing import Optional, Dict, Any, List, Union, Tuple, TYPE_CHECKING

from overrides import overrides
import torch

from allennlp.common import Params
from allennlp.training.callbacks.callback import TrainerCallback
from allennlp.training.callbacks.log_writer import LogWriterCallback


if TYPE_CHECKING:
    from allennlp.training.trainer import GradientDescentTrainer


@TrainerCallback.register("wandb")
class WandBCallback(LogWriterCallback):
    """
    Logs training runs to Weights & Biases.

    !!! Note
        This requires the environment variable 'WANDB_API_KEY' to be set.

    In addition to the parameters that `LogWriterCallback` takes, there are several other
    parameters specific to `WandBWriter` listed below.

    # Parameters

    project : `Optional[str]`, optional (default = `None`)
        The name of the W&B project to save the training run to.
    tags : `Optional[List[str]]`, optional (default = `None`)
        Tags to assign to the training run in W&B.
    watch_model : `bool`, optional (default = `True`)
        Whether or not W&B should watch the `Model`.
    files_to_save : `Tuple[str, ...]`, optional (default = `("config.json", "out.log")`)
        Extra files in the serialization directory to save to the W&B training run.
    """

    def __init__(
        self,
        serialization_dir: str,
        summary_interval: int = 100,
        distribution_interval: Optional[int] = None,
        batch_size_interval: Optional[int] = None,
        should_log_parameter_statistics: bool = True,
        should_log_learning_rate: bool = False,
        project: Optional[str] = None,
        tags: Optional[List[str]] = None,
        watch_model: bool = True,
        files_to_save: Tuple[str, ...] = ("config.json", "out.log"),
    ) -> None:
        if "WANDB_API_KEY" not in os.environ:
            raise ValueError("Missing environment variable 'WANDB_API_KEY'")

        super().__init__(
            serialization_dir,
            summary_interval=summary_interval,
            distribution_interval=distribution_interval,
            batch_size_interval=batch_size_interval,
            should_log_parameter_statistics=should_log_parameter_statistics,
            should_log_learning_rate=should_log_learning_rate,
        )

        self._watch_model = watch_model
        self._files_to_save = files_to_save

        import wandb

        self.wandb = wandb
        self.wandb.init(
            dir=os.path.abspath(serialization_dir),
            project=project,
            config=Params.from_file(os.path.join(serialization_dir, "config.json")).as_dict(),
            tags=tags,
        )

        for fpath in self._files_to_save:
            self.wandb.save(os.path.join(serialization_dir, fpath), base_path=serialization_dir)

    @overrides
    def log_scalars(
        self,
        scalars: Dict[str, Union[int, float]],
        log_prefix: str = "",
        epoch: Optional[int] = None,
    ) -> None:
        self._log(scalars, log_prefix=log_prefix, epoch=epoch)

    @overrides
    def log_tensors(
        self, tensors: Dict[str, torch.Tensor], log_prefix: str = "", epoch: Optional[int] = None
    ) -> None:
        self._log(
            {k: self.wandb.Histogram(v.cpu().data.numpy().flatten()) for k, v in tensors.items()},
            log_prefix=log_prefix,
            epoch=epoch,
        )

    def _log(
        self, dict_to_log: Dict[str, Any], log_prefix: str = "", epoch: Optional[int] = None
    ) -> None:
        if log_prefix:
            dict_to_log = {f"{log_prefix}/{k}": v for k, v in dict_to_log.items()}
        if epoch is not None:
            dict_to_log["epoch"] = epoch
        self.wandb.log(dict_to_log, step=self.trainer._batch_num_total)  # type: ignore[union-attr]

    @overrides
    def on_start(
        self, trainer: "GradientDescentTrainer", is_primary: bool = True, **kwargs
    ) -> None:
        super().on_start(trainer, is_primary=is_primary, **kwargs)
        if self._watch_model:
            self.wandb.watch(self.trainer.model)  # type: ignore[union-attr]
